from ..object_compare_tracking import ObjectComparisonTracking

import enstools.feature.identification.african_easterly_waves.configuration as cfg
from datetime import datetime, timedelta

from shapely.geometry import Polygon, LineString
from enstools.feature.util.data_utils import pb_str_to_datetime


class AEWTracking(ObjectComparisonTracking):
    """
    Tracking of AEWs.
    Based on comparisons on wavetroughs
    """

    def __init__(self, **kwargs):
        self.config = cfg
        self.set_max_delta_compare(cfg.max_cmp_delta)
        pass

    def correspond(self, time1, obj1, time2, obj2):
        # returns True if objects are same wave

        # access the properties set earlier
        prop1 = obj1.properties
        prop2 = obj2.properties
        lineseg1 = prop1.line_pts
        lineseg2 = prop2.line_pts

        bb1_lon_center = (prop1.bb.max.lon - prop1.bb.min.lon) / 2.0
        bb2_lon_center = (prop2.bb.max.lon - prop2.bb.min.lon) / 2.0

        time_diff = time2 - time1
        time_diff_h = time_diff.seconds // 3600

        # predicted polygonal area
        predicted_polygon_list = []
        for i in range(len(lineseg1)):  # min speed
            predicted_polygon_list.append(
                [lineseg1[i].lon + self.config.speed_deg_per_h[0] * time_diff_h, lineseg1[i].lat])
        for i in range(len(lineseg1) - 1, -1, -1):  # max speed
            predicted_polygon_list.append(
                [lineseg1[i].lon + self.config.speed_deg_per_h[1] * time_diff_h, lineseg1[i].lat])
        predicted_polygon = Polygon(predicted_polygon_list)

        # actual line string
        p2_pt_list = []
        for i in range(len(lineseg2)):
            p2_pt_list.append((lineseg2[i].lon, lineseg2[i].lat))
        actual_linestring = LineString(p2_pt_list)

        if not predicted_polygon.intersects(actual_linestring): # linesting poly not where expected?
            return False

        # if bb1_lon_center - bb2_lon_center  > time_diff_h / 2: # TODO weird
        #     return False

        return True

    def postprocess(self, obj_desc):
        print("postprocess @ aew")
        return

    def adjust_track(self, track):
        # keep track if persists longer than duration_threshold

        nodes = [edge.parent for edge in track.edges]
        min_time = pb_str_to_datetime(nodes[0].time)
        max_time = pb_str_to_datetime(nodes[-1].time)
        duration = max_time - min_time

        if duration < self.config.duration_threshold:
            return None

        times = [pb_str_to_datetime(node.time) for node in nodes]
        center_lons = [(node.object.properties.bb.max.lon + node.object.properties.bb.min.lon) / 2.0 for node in nodes]
        center_lats = [(node.object.properties.bb.max.lat + node.object.properties.bb.min.lat) / 2.0 for node in nodes]

        # TODO need general statement that nodes are part of exactly one track.

        # init nodes as
        keep_node = [False] * len(nodes)

        for node_idx, node in enumerate(nodes):
            # TODO v this should instantly return pairs (endnote, all nodes on path) for later
            downstream_nodes = get_downstream_nodes_after(self.config.slow_duration_window, node_idx, track.edges)
            # TODO if none? maybe skipped?

            # for each +48h node: check if avg speed fulfilled. if yes, mark this path.
            for ds in downstream_nodes:
                ds_node_lon_center = (ds.object.properties.bb.max.lon + ds.object.properties.bb.min.lon) / 2.0

                deg_per_hr_avg = (ds_node_lon_center - center_lons[node_idx]) / self.config.slow_duration_window.hours
                if abs(deg_per_hr_avg) > abs(self.config.avg_speed_thr):
                    # fast enough, keep all nodes along this path.
                    keep_idxs = get_node_indices_between(node, ds, nodes)
                    for keep_idx in keep_idxs:
                        keep_node[keep_idx] = True

        kept_nodes = [node for x, node in enumerate(nodes) if keep_node[x]]

        new_tracks = generate_tracks(kept_nodes) # regenerate tracks out of leftover edges.

        return new_tracks

        #print(times)
        #print(center_lons)
        #exit()
        """
        for x,x+48h
            center point dist in deg?
            if avg_48_speed < avg_speed_thr:
                remove pts in between
                recursive on two leftover tracks?
        """
        # TODO for each 48h window:
        #   didnt move far enough -> trim start/end or split
        #   redo clim, plot which WTs not considered

        return track
